/*
  Copyright 2005 Damien Stehlé.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.

  This program implements ideas from the paper "Gal's Accurate Tables Method 
  Revisited", by Damien Stehlé and Paul Zimmermann, in the Proceedings of 
  Arith'17, IEEE.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <gmp.h>
#include <mpfr.h>

#include "mat.h"
#include "basicfast.h"
#include "fast.h"
#include "gatmr.h" 

#define VERSION "1.0"
// #define VERBOSE
#define MALLOC malloc
#define FREE free


/* return the degree of r[0]+t*r[1]+...+t^n*r[n], -1 if identically 0 */
int
Degree (mpz_t *r, int n)
{
  while (n >= 0 && mpz_cmp_ui (r[n], 0) == 0)
    n--;
  return n;
}


void
Res22 (mpz_t *r, mpz_t *a, mpz_t *b, mpz_t *c, mpz_t tmp)
{
  /* r1:=-(a3b4)c2+(a3c4)b2+(c3b4)a2-(a4c3)b2+(a4b3)c2-(b3c4)a2 */
  /* r0:=-(a3b4)c1+(a3c4)b1+(c3b4)a1-(a4c3)b1+(a4b3)c1-(b3c4)a1 */
  /* compute first the common products */

  mpz_mul (tmp,  a[3], b[4]);   /* tmp=a3b4 */
  mpz_mul (a[3], a[3], c[4]);   /* a3 =a3c4 */
  mpz_mul (b[4], c[3], b[4]);   /* b4 =c3b4 */
  mpz_mul (c[3], a[4], c[3]);   /* c3 =a4c3 */ 
  mpz_mul (a[4], b[3], a[4]);   /* a4 =b3a4 */
  mpz_mul (c[4], c[4], b[3]);   /* c4 =c4b3 */

  mpz_mul (r[0], tmp, c[1]);
  mpz_mul (b[3], a[3], b[1]);
  mpz_sub (r[0], b[3], r[0]);
  mpz_mul (b[3], a[1], b[4]); 
  mpz_add (r[0], r[0], b[3]);
  mpz_mul (b[3], c[3], b[1]);
  mpz_sub (r[0], r[0], b[3]);
  mpz_mul (b[3], a[4], c[1]);
  mpz_add (r[0], r[0], b[3]);
  mpz_mul (b[3], a[1], c[4]);
  mpz_sub (r[0], r[0], b[3]);  
  /*  mpz_mod (r[0], r[0], q); */

  mpz_mul (r[1], tmp, c[2]);
  mpz_mul (b[3], a[3], b[2]);
  mpz_sub (r[1], b[3], r[1]);
  mpz_mul (b[3], a[2], b[4]); 
  mpz_add (r[1], r[1], b[3]);
  mpz_mul (b[3], c[3], b[2]);
  mpz_sub (r[1], r[1], b[3]);
  mpz_mul (b[3], a[4], c[2]);
  mpz_add (r[1], r[1], b[3]);
  mpz_mul (b[3], a[2], c[4]);
  mpz_sub (r[1], r[1], b[3]);  
  /*  mpz_mod (r[0], r[0], q); */
}



/* INITIALIZATION OF THE MATRIX TO BE REDUCED */


void 
set_matrix (mat_ZZ *B, mpz_t T, mpz_t C, mpz_t a1, mpz_t a2, mpz_t b1,
            mpz_t b2, mpz_t c1, mpz_t c2)
{
  mpz_t **L = B->coeff;
  int i;

  mpz_set (L[1][1], C);                    /* L11=C */
  for (i=2; i<=5; i++)
    mpz_set_ui (L[1][i], 0);

  mpz_set_ui (L[2][1], 0);
  mpz_mul (L[2][2], L[1][1], T);           /* L22=CT */   
  for (i=3; i<=5; i++)
    mpz_set_ui (L[2][i], 0);

  mpz_set (L[3][1], a1);                   /* L31=a1 */
  mpz_set (L[3][2], b1);                   /* L32=b1 */
  mpz_set (L[3][3], c1);                   /* L33=c1 */
  mpz_set_ui (L[3][4], 3);                 /* L34=3 */
  mpz_set_ui (L[3][5], 0);

  mpz_set (L[4][1], a2);                   /* L41=a2 */
  mpz_set (L[4][2], b2);                   /* L42=b2 */
  mpz_set (L[4][3], c2);                   /* L43=c2 */
  mpz_set_ui (L[4][4], 0);
  mpz_set_ui (L[4][5], 3);                 /* L45=3 */

#ifdef VERBOSE
  printf ("Set matrix is:\n");  Print_mat(B->coeff, 4, 5); 
#endif
}


/* tmp <- abs(V[1])+...+abs(V[5]) */
void
Norm1 (mpz_t tmp, mpz_t *V)
{
  unsigned long i;

  mpz_set_ui (tmp, 0);

  for (i=1; i<=5; i++)
    {
      if (mpz_cmp_ui (V[i], 0) >= 0)
        mpz_add (tmp, tmp, V[i]);
      else
        mpz_sub (tmp, tmp, V[i]);
    }
}



/* COMPUTES THE INTEGER LINEAR COEFFICIENTS OF V AS REGARD TO THE 4 INITIAL
   LATTICE VECTORS, AND STORE IT IN V. */

void
get_lin_coeffs (mpz_t *V, mpz_t a1, mpz_t a2, mpz_t b1, mpz_t b2, mpz_t C, 
		mpz_t T, mpz_t tmp)
{
  /*printf ("V="); 
 mpz_out_str (stdout, 10, V[1]); printf ("\n");
 mpz_out_str (stdout, 10, V[2]); printf ("\n");
 mpz_out_str (stdout, 10, V[3]); printf ("\n");
 mpz_out_str (stdout, 10, V[4]); printf ("\n");
 mpz_out_str (stdout, 10, V[5]); printf ("\n");*/

  /* a3 := V[4]/3 */
  mpz_divexact_ui (V[3], V[4], 3);

  /* a4 := v[5]/3 */
  mpz_divexact_ui (V[4], V[5], 3);

  /* a2 := (v[2]-a3*b1 -a4*b2)/(CT) */
  mpz_mul (tmp, b1, V[3]);
  mpz_sub (V[2], V[2], tmp);
  mpz_mul (tmp, b2, V[4]);
  mpz_sub (V[2], V[2], tmp);
  mpz_mul (tmp, C, T);
  mpz_divexact (V[2], V[2], tmp);


  /* a1 := (v[1]-a2*CT-a3*a1-a4*a2)/(C) */
  mpz_mul (tmp, a1, V[3]);
  mpz_sub (V[1], V[1], tmp);
  mpz_mul (tmp, a2, V[4]);
  mpz_sub (V[1], V[1], tmp);
  mpz_divexact (V[1], V[1], C);

}



/* Res[i] <- Res[i] - q*round(Res[i]/q), where e = (q-1)/2 */
void
centermod (mpz_t *Res, int d, mpz_t q)
{
  mpz_t e;
  int i;

  mpz_init (e);
  mpz_sub_ui (e, q, 1);
  mpz_div_2exp (e, e, 1); /* e = (q-1)/2 */
  for (i=0; i<d; i++)
    {
      mpz_mod (Res[i], Res[i], q);
      if (mpz_cmpabs (Res[i], e) > 0)
        mpz_sub (Res[i], Res[i], q);
    }
  mpz_clear (e);
}


/* M = 2^m, T = 2^t
   Return -1 in case of failure, otherwise number of worst cases found.
 */


int
slz (mpz_t *Res, mpfr_t *P1, mpfr_t *P2, mpz_t T, mpz_t q, mpz_t C, mat_ZZ *B)
{
  int i, j, k;
  mpz_t a1, a2, b1, b2, c1, c2, tmp;
  int ok[5], dRes;

  mpz_init (a1);
  mpz_init (a2);
  mpz_init (b1);
  mpz_init (b2);
  mpz_init (c1);
  mpz_init (c2);
  mpz_init (tmp);

  /* a1,2 := round(C*coeff(P_(1,2),t,0)) */
  mpfr_mul_z (P1[0], P1[0], C, GMP_RNDN);
  mpfr_round (P1[0], P1[0]);
  mpfr_get_z (a1, P1[0], GMP_RNDN);
  mpfr_mul_z (P2[0], P2[0], C, GMP_RNDN);
  mpfr_round (P2[0], P2[0]);
  mpfr_get_z (a2, P2[0], GMP_RNDN);

  /* a := a - centered_round(C/a)*C */
  mpz_tdiv_q (b1, a1, C);
  mpz_mul (b1, b1, C);
  mpz_sub (a1, a1, b1);
  mpz_mul_2exp (b1, a1, 1);
  if (mpz_cmp (b1, C) >= 0)
    mpz_sub (a1, a1, C);
  mpz_tdiv_q (b2, a2, C);
  mpz_mul (b2, b2, C);
  mpz_sub (a2, a2, b2);
  mpz_mul_2exp (b2, a2, 1);
  if (mpz_cmp (b2, C) >= 0)
    mpz_sub (a2, a2, C);

  /* b12 := round(C*coeff(P12,t,1)*T) */
  mpfr_mul_z (P1[1], P1[1], C, GMP_RNDN);
  mpfr_mul_z (P1[1], P1[1], T, GMP_RNDN);
  mpfr_round (P1[1], P1[1]);
  mpfr_get_z (b1, P1[1], GMP_RNDN);
  mpfr_mul_z (P2[1], P2[1], C, GMP_RNDN);
  mpfr_mul_z (P2[1], P2[1], T, GMP_RNDN);
  mpfr_round (P2[1], P2[1]);
  mpfr_get_z (b2, P2[1], GMP_RNDN);

  /* c12 := round(C*coeff(P12,t,2)*T2 */
  mpfr_mul_z (P1[2], P1[2], C, GMP_RNDN);
  mpfr_mul_z (P1[2], P1[2], T, GMP_RNDN);
  mpfr_mul_z (P1[2], P1[2], T, GMP_RNDN);
  mpfr_round (P1[2], P1[2]);
  mpfr_get_z (c1, P1[2], GMP_RNDN);
  mpfr_mul_z (P2[2], P2[2], C, GMP_RNDN);
  mpfr_mul_z (P2[2], P2[2], T, GMP_RNDN);
  mpfr_mul_z (P2[2], P2[2], T, GMP_RNDN);
  mpfr_round (P2[2], P2[2]);
  mpfr_get_z (c2, P2[2], GMP_RNDN);

#ifdef VERBOSE
  printf ("a1="); mpz_out_str (stdout, 10, a1); printf ("\n");
  printf ("a2="); mpz_out_str (stdout, 10, a2); printf ("\n");
  printf ("b1="); mpz_out_str (stdout, 10, b1); printf ("\n");
  printf ("b2="); mpz_out_str (stdout, 10, b2); printf ("\n");
  printf ("c1="); mpz_out_str (stdout, 10, c1); printf ("\n");
  printf ("c2="); mpz_out_str (stdout, 10, c2); printf ("\n");
#endif

  /*STEP 6 OF THE SLZ ALGORITHM */

  set_matrix (B, T, C, a1, a2, b1, b2, c1, c2);

#ifdef VERBOSE
  printf ("\nInput matrix:\n"); Print_mat (B->coeff, 4, 5); 
#endif

  /*STEP 7 OF THE SLZ ALGORITHM, LATTICE REDUCTION */

  LLL (B);

#ifdef VERBOSE  
  printf ("Computed matrix B is:\n");  Print_mat (B->coeff, 4, 5);
#endif


  /* TAKE THE 3 SHORTEST VECTORS THAT ARE ALG. INDEPENDENT */
  for (i=1; i<=4; i++) ok[i] = -1;
  dRes= -1;

  for (i = 1; i <= 4 && dRes == -1; i++)
    {
      if (ok[i] == -1)
        {
          Norm1 (tmp, B->coeff[i]);
          if ((ok[i]=(mpz_cmp (tmp, C) < 0)))
            get_lin_coeffs (B->coeff[i], a1, a2, b1, b2, C, T, tmp);
        }
      if (ok[i])
        {
          for (j = i + 1; j <= 4 && dRes == -1; j++)
            {
              if (ok[j] == -1)
                {
                  Norm1 (tmp, B->coeff[j]);
                  if ((ok[j] = (mpz_cmp (tmp, C) < 0)))
                    get_lin_coeffs (B->coeff[j], a1, a2, b1, b2, C, T, tmp);
                }

              if (ok[j])
                {
		  for (k = j + 1; k <= 4 && dRes == -1; k++)
		    {
		      if (ok[k] == -1)
			{
			  Norm1 (tmp, B->coeff[k]);
			  
			  //printf("COUCOU %d\n",  (mpz_cmp (tmp, C) < 0));

			  if ((ok[k] = (mpz_cmp (tmp, C) < 0)))
			    get_lin_coeffs (B->coeff[k], a1, a2, b1, b2, C, T, tmp);
			}
		      if (ok[k]>=1)
			{
#ifdef VERBOSE
			  printf("i=%d, j=%d, k=%d, ok[k]=%d \n", i, j, k, ok[k]);
			  printf ("Computed matrix B is:\n");  
			  Print_mat(B->coeff, 4, 5); 
#endif
			  
			  Res22 (Res, B->coeff[i], B->coeff[j], B->coeff[k], tmp);
			  
#ifdef VERBOSE
			  printf("Resultant equals: "); 
			  Print_vec (Res, 1); printf("\n");
#endif
			  
			  dRes = Degree (Res, 1);
			}
		    }
		}
	    }
	}
    }
  
#ifdef VERBOSE
  if (dRes == -1)
    fprintf (stderr, "Error, no valid vector found\n");
#endif



  /* Compute the roots of the resultant, that's where q becomes useful */
  if (dRes > 0)
    {
      mpz_invert (Res[1], Res[1], q);
      mpz_mul (Res[0], Res[0], Res[1]);
      mpz_neg (Res[0], Res[0]); 
      centermod (Res, 1, q);

#ifdef VERBOSE
      printf ("centered Roots are "); 
      mpz_out_str (stdout, 10, Res[0]); printf ("\n");

#endif
    }

  mpz_clear (a1);
  mpz_clear (b1);
  mpz_clear (c1);
  mpz_clear (a2);
  mpz_clear (b2);
  mpz_clear (c2);
  mpz_clear (tmp);
  return dRes;
}





/* checks that the retuned d possible worst-cases stored in vector Res
   satisfy  centered_frac( 2N*sin,cos[(Res[i]+t0)/N]  < 1/M ) 
   returns the number of real bad cases, k*/

int
Check (mpz_t *Res, int d, mpz_t t0, int n, int nn, mpfr_t oneoverM)
{
  int i, k = 0;
  mpfr_t x1, x2, y1, y2;

  mpfr_init (x1);
  mpfr_init (x2);
  mpfr_init (y1);
  mpfr_init (y2);
  

  for (i=0; i<d; i++)
    {
    /*compute cfrac(NN*sin,cos[ (Res[i]+t0)/N] )   */
      mpz_add (Res[i], Res[i], t0);
      mpfr_set_z (x1, Res[i], GMP_RNDN);
      mpfr_div_2si (x1, x1, n, GMP_RNDN);
      mpfr_sin_cos (x1, x2, x1, GMP_RNDN);    
      mpfr_mul_2si (x1, x1, nn , GMP_RNDN);
      mpfr_mul_2si (x2, x2, nn , GMP_RNDN);
      mpfr_round (y1, x1);
      mpfr_round (y2, x2);
      mpfr_sub (y1, x1, y1, GMP_RNDN);
      mpfr_sub (y2, x2, y2, GMP_RNDN);
      mpfr_abs (y1, y1, GMP_RNDN);
      mpfr_abs (y2, y2, GMP_RNDN);

    /*check if x1,x2 < 1/2^m */
      if ( (mpfr_cmp_ui (y1, 0) == 0) || (mpfr_cmp_ui (y2, 0) == 0) )
        {
          fprintf (stderr, "Error, precision is too small\n"); exit (1);
        }
      
#ifdef VERBOSE
      printf("oneoverM= "); mpfr_out_str (stdout, 10, 10, oneoverM, GMP_RNDN); 
      printf ("\n");
      printf("y1= "); mpfr_out_str (stdout, 10, 10, y1, GMP_RNDN); 
      printf ("\n");
      printf("y2= "); mpfr_out_str (stdout, 10, 10, y2, GMP_RNDN); 
      printf ("\n");
#endif


      if ( (mpfr_cmp (y1, oneoverM) <= 0) && (mpfr_cmp (y2, oneoverM) <= 0) )
        {
	  /* IMPRESSION DU PIRE CAS TROUVE */
          k ++;

	  /*
          printf ("*** x=");
          mpz_out_str (stdout, 16, Res[i]);
          printf (" m=%lu,%lu\n", -mpfr_get_exp (y1),-mpfr_get_exp (y2)); */
	  fprintf (stderr, "*** x=");
          mpz_out_str (stderr, 10, Res[i]);
          fprintf (stderr, " m=%lu,%lu\n", -mpfr_get_exp (y1),-mpfr_get_exp (y2));

        }
    }
  mpfr_clear (x1);
  mpfr_clear (x2);
  mpfr_clear (y1);
  mpfr_clear (y2);

  return k;
}




/**************************************************************************/


/* Usage: wclr prec n m t t0 t1
   Prints all worst-cases at distance < M for ->>> sin and cos <<<- 
and t0 <= t <= t1,
   by intervals of length q=2T+1, with T=2^t, where N=2^n, M=2^m.
*/

int
main (int argc, char *argv[])
{
  mp_prec_t prec;
  int n, nn, number = 0;
  double m, t, initial_t;
  mpz_t t0, t1, T, q, *Res, tm, C, total_T;
  mpfr_t P1[3];
  mpfr_t P2[3];
  mpfr_t x1, x2;
  mpfr_t tmp, oneoverM;
  double eps, Mprime;
  int d, verbose = 0, first = 0;
  int tries = 0;              /* number of calls to slz */
  int failures = 0;           /* number of failures */
  mat_ZZ B;

  fprintf (stderr, "\nWelcome to gatmr version %s\n", VERSION);


  if (argc > 1 && strcmp (argv[1], "-v") == 0)
    {
      verbose = 1;
      argv ++;
      argc --;
    }

  if (argc > 1 && strcmp (argv[1], "-first") == 0)
    {
      first = 1;
      argv ++;
      argc --;
    }

  if (argc != 8)
    {
      fprintf (stderr, "Usage: ./gal [-v] [-first] prec n nn m t t0 t1\n");
      exit (1);
    }

  init_matrix (&B, 4, 5);



  /* Initializing n, nn, m, t, initial_t, t0, t1, tm, T=2^t, q=2T+1 */

  prec = atoi (argv[1]);
  mpfr_set_default_prec (prec);


  n = atoi (argv[2]);
  nn = atoi (argv[3]);
  m = atof (argv[4]);
  t = atof (argv[5]);
  initial_t = t;

  mpz_init (t0);
  mpz_init (t1);
  mpz_init (tm);
  mpz_init (T);
  mpz_init (q);
  mpz_init (C);
  mpfr_init (tmp);
  mpfr_init (oneoverM);
  mpz_init (total_T);

  mpz_set_ui (total_T, 0);

  mpz_set_str (t0, argv[6], 10);
  mpz_set_str (t1, argv[7], 10);


  /* oneoverM = 2^-m */
  mpfr_set_d (tmp, -m, GMP_RNDN);
  mpfr_ui_pow (oneoverM, 2, tmp, GMP_RNDN);


  /* T = 2^t */
  mpfr_set_d (tmp, t, GMP_RNDN);
  mpfr_ui_pow (tmp, 2, tmp, GMP_RNDN);
  mpfr_get_z (T, tmp, GMP_RNDN);

  /* q = nextprime(2*T) */
  mpz_mul_2exp (q, T, 1);
  mpz_nextprime (q, q);

  fprintf (stderr, 
	   "You are looking for %e-simultaneously bad cases for the \nrounding of cos(x) and sin(x), over [%e,%e] with:\n -input precision: %d bits\n -output precision: %d bits\n -working precision: %d bits\n -size of subintervals: %e\n \n", 
	   m, 
	   (mpz_get_d (t0))/ exp(log(2)*((double) n)), 
	   (mpz_get_d (t1))/ exp(log(2)*((double) n)), 
	   n, nn, atoi (argv[1]), 
	   ((double) exp(log(2)*(t-((double) n)))));


#ifdef VERBOSE
  printf ("q="); mpz_out_str (stdout, 10, q); printf ("\n");
#endif

  /* with Mac Laurin's formula, we get that eps<=NN*T^3/(6N^3) */

  mpfr_init (x1); 
  mpfr_init (x2); 
  mpfr_init (P1[0]);
  mpfr_init (P1[1]);
  mpfr_init (P1[2]);
  mpfr_init (P2[0]);
  mpfr_init (P2[1]);
  mpfr_init (P2[2]);

  mpfr_set_ui (tmp, 1, GMP_RNDN);
  mpfr_div_2si (x1, tmp, 3 * n - nn , GMP_RNDU);
  mpfr_set_z (tmp, T, GMP_RNDN);
  mpfr_pow_ui (tmp, tmp, 3, GMP_RNDZ);
  mpfr_div (x1, x1, tmp, GMP_RNDU);
  mpfr_div_ui (x1, x1, 6, GMP_RNDU);
  eps = mpfr_get_d (x1, GMP_RNDU);

  
#ifdef VERBOSE
  printf("eps= "); mpfr_out_str (stdout, 10, 10, x1, GMP_RNDN); printf ("\n");
#endif

  /* Mprime := floor((1/2)/[1/M+eps]) */

  mpfr_set_d (tmp, m, GMP_RNDN);
  mpfr_ui_pow (tmp, 2, tmp, GMP_RNDZ);
  Mprime = mpfr_get_d (tmp, GMP_RNDZ);          /* M'=2^m */
  eps += 1 / Mprime ;
  Mprime = floor (0.5 / eps);

#ifdef VERBOSE
  printf ("eps=%1.20e M'=%1.0f\n", eps, Mprime);
#endif
  
  /* C:=3*Mprime */

  mpz_set_d (C, 3.0*Mprime);

#ifdef VERBOSE
  printf ("C="); mpz_out_str (stdout, 10, C); printf ("\n");
#endif




  /* THE NICE PART OF THE ALGORITHM: MULTI-CALLS TO SLZ */
  /* We prepare the place for only 2 bad cases, since the resultant has 
     degree 2, thus it has at most 2 roots. */

  Res = init_vec (2);
  
  while (mpz_cmp (t0, t1) < 0)
    {
      if (verbose)
	{
	  printf ("x="); mpz_out_str (stdout, 10, t0); printf ("\n");
	}

      mpz_add (tm, t0, T);        /* tm is the middle of current interval */
      
      /*INITIALIZATION OF x=f(tm) AND APPROX. POLYNOMIALS P1 AND P2*/ 

      /* compute x1 = sin(tm/N), x2 = cos(tm/N)*/ 
      mpfr_set_z (x1, tm, GMP_RNDN);
      mpfr_div_2si (x1, x1, n, GMP_RNDN);
      mpfr_sin_cos (x1, x2, x1, GMP_RNDN);

      /* compute series expansion 
         P1(x)= sin(xm) + cos(xm)(x-xm) -sin(xm)/2 (x-xm)^2
	 P2(x)= cos(xm) - sin(xm)(x-xm) -cos(xm)/2 (x-xm)^2 */
      mpfr_mul_2si (P1[0], x1, nn, GMP_RNDN);             /* NN*f(tm) */
      mpfr_mul_2si (P2[0], x2, nn, GMP_RNDN);
      mpfr_mul_2si (P1[1], x2, nn-n, GMP_RNDN);           /* NN/N * f'(tm) */
      mpfr_neg (P2[1], x1, GMP_RNDN);
      mpfr_mul_2si (P2[1], P2[1], nn-n, GMP_RNDN); 
      mpfr_neg (P1[2], x1, GMP_RNDN);                     /* NN*f"(tm)/N^2 */
      mpfr_div_2si (P1[2], P1[2], 2*n+1-nn, GMP_RNDN); 
      mpfr_neg (P2[2], x2, GMP_RNDN);
      mpfr_div_2si (P2[2], P2[2], 2*n+1-nn, GMP_RNDN);      

#ifdef VERBOSE
          printf ("\nP0=");
	  mpfr_out_str (stdout, 10, 10, P1[0], GMP_RNDN); printf ("  ");
	  mpfr_out_str (stdout, 10, 10, P2[0], GMP_RNDN); printf ("\n");
          printf ("P1=");
          mpfr_out_str (stdout, 10, 10, P1[1], GMP_RNDN); printf ("  ");
	  mpfr_out_str (stdout, 10, 10, P2[1], GMP_RNDN); printf ("\n");
          printf ("P2=");
          mpfr_out_str (stdout, 10, 10, P1[2], GMP_RNDN); printf ("  ");
	  mpfr_out_str (stdout, 10, 10, P2[2], GMP_RNDN); printf ("\n"); 
	  printf("\n"); 
#endif

      d = slz (Res, P1, P2, T, q, C, &B);
      
      tries ++;
      if (d == -1)
        {
          failures ++;

	  /* In case SLZ fails, we divide the length of the interval by 2,
	     and we start again. */

          if (verbose || t < 1.0)
            {
              fprintf (stderr, "FAIL at x="); mpz_out_str (stderr, 10, t0);
              fprintf (stderr, " with T="); mpz_out_str (stderr, 10, T);
              fprintf (stderr, "\n");
              if (t < 1.0) exit (1);
            }
          t = t - 1.0;
	  mpfr_set_d (tmp, t, GMP_RNDN);
	  mpfr_ui_pow (tmp, 2, tmp, GMP_RNDN);
	  mpfr_get_z (T, tmp, GMP_RNDN);
        }
      else
        {
	  /* update total_T */
	  mpz_add (total_T, total_T, T);
	  mpz_add (total_T, total_T, T);
 
         /* check potential solutions and print them if they are valid.*/

	  
	  /*
	  if (Check (Res, d, tm, n, nn, oneoverM) && first) break;
	  */

	  number += Check (Res, d, tm, n, nn, oneoverM);

          mpz_add (t0, tm, T);
	  
	  /* If t is smaller than initial_t, we reset t and T to their
	     initial values. */
	  
          if (t < initial_t)
            {
              t = initial_t;
	      mpfr_set_d (tmp, t, GMP_RNDN);
	      mpfr_ui_pow (tmp, 2, tmp, GMP_RNDN);
	      mpfr_get_z (T, tmp, GMP_RNDN);
            }
        }
    }
  

  /* DESCRIPTION OF THE BEHAVIOR OF THE ALGORITHM */

  /* ATTENTION, THE "average width" BELOW TAKES INTO ACCOUNT THE CASES 
     WHERE IT FAILS AND THE INTERVAL LENGTH HAD TO BE SHRUNK. */

  mpfr_set_z (tmp, total_T, GMP_RNDN);
  mpfr_div_ui (tmp, tmp, tries, GMP_RNDN);

  fprintf (stderr, 
	   "%d lattices, %d failures, %.2lf failure rate, average width ",
          tries, failures, ((double) failures)/((double) tries));
  mpfr_out_str (stderr, 10, 3, tmp, GMP_RNDN); fprintf (stderr, "\n");
  fprintf (stderr, "You got %d solutions.\n", number);

  clear_vec (Res, 2);
  
  mpfr_clear (tmp);
  mpfr_clear (x1);
  mpfr_clear (x2);
  mpfr_clear (P1[0]);
  mpfr_clear (P1[1]);
  mpfr_clear (P1[2]);
  mpfr_clear (P2[0]);
  mpfr_clear (P2[1]);
  mpfr_clear (P2[2]);
  mpfr_clear (oneoverM);

  mpz_clear (t0);
  mpz_clear (t1);
  mpz_clear (tm);
  mpz_clear (T);
  mpz_clear (q);
  mpz_clear (C);
  mpz_clear (total_T);
  clear_matrix (&B, 4, 5);
  return 0;
}


